# Copyright 2023, European Centre for Medium Range Weather Forecasts.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
import warnings

import matplotlib
import numpy as np
from matplotlib.colors import BoundaryNorm, LinearSegmentedColormap

from earthkit.maps.schema import schema

MAPPED_COLORS = {
    "mask": (0, 0, 0, 0),
}


def colormap(cmap, levels):
    n_colors = len(levels) - 1

    if not isinstance(cmap, (list, tuple)):
        colors = [matplotlib.cm.get_cmap(cmap)(i) for i in np.linspace(0, 1, n_colors)]
    else:
        colors = [read_color(color) for color in cmap]
    colormap = LinearSegmentedColormap.from_list(name="", colors=colors, N=n_colors)
    colormap.set_under((0, 0, 0, 0))

    norm = BoundaryNorm(levels, colormap.N)

    return colormap, norm


def read_color(color):
    if isinstance(color, str):
        if color.lower() == "none":
            color = "#ffffff"
        elif color.lower().startswith("rgb"):
            result = re.search("\(([^)]+)", color)  # noqa: W605
            if result is None:
                raise ValueError(f"unrecognised color {color}")
            else:
                color = tuple(float(x) for x in result.group(1).split(","))
    return color


def dynamic(normalize=False):
    def decorator(method):
        def wrapper(self, *args, **kwargs):
            if "colors" in kwargs:
                style = kwargs.pop("style", None)
                if style is not None:
                    warnings.warn(
                        f"Both 'colors' and 'style' passed to {method.__name__}; "
                        f"using 'colors' instead of 'style'"
                    )
                kwargs["cmap"] = kwargs.pop("colors")
            else:
                kwargs["cmap"] = kwargs.pop(
                    "style", kwargs.pop("cmap", schema.default_style)
                )

            if normalize and "levels" in kwargs:
                cmap, norm = colormap(kwargs.pop("cmap"), kwargs["levels"])
                kwargs.update({"cmap": cmap, "norm": norm})

            color_bounds = []
            for kwarg in ("under_vmin", "over_vmax"):
                breach_color = kwargs.pop(kwarg, None)
                if breach_color is not None:
                    direction, threshold_key = kwarg.split("_")
                    threshold_value = kwargs.get(threshold_key)
                    if threshold_value is None:
                        raise ValueError(
                            f"'{kwarg}' can only be passed if '{threshold_key}' "
                            f"is also passed"
                        )
                    breach_color = MAPPED_COLORS.get(breach_color, breach_color)
                    color_bounds.append((f"set_{direction}", breach_color))

            result = method(self, *args, **kwargs)
            layer = self._layers[-1].layer

            for function, arg in color_bounds:
                getattr(layer.cmap, function)(arg)

            return result

        return wrapper

    return decorator



# Copyright 2023, European Centre for Medium Range Weather Forecasts.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import glob
import os

from matplotlib import font_manager, rcParams

from earthkit.maps._definitions import FONTS_DIR
from earthkit.maps.schema import schema


def register_fonts():
    fontpaths = glob.glob(os.path.join(FONTS_DIR, "*"))
    for fontpath in fontpaths:
        font_files = glob.glob(os.path.join(fontpath, "*.ttf"))
        for font_file in font_files:
            font_manager.fontManager.addfont(font_file)


register_fonts()
rcParams["font.family"] = schema.font



# Copyright 2023, European Centre for Medium Range Weather Forecasts.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import warnings

import cartopy.crs as ccrs
import cv2
import emohawk
import numpy as np

from earthkit.maps import domains
from earthkit.maps.schema import schema

# from cartopy.util import add_cyclic_point


X_RESOLUTION = 1000
Y_RESOLUTION = 1000

DEFAULT_PROJ = ccrs.PlateCarree()

COMMON_VARIABLES = {
    "u": ["u", "u10"],
    "v": ["v", "v10"],
}


def find(values, candidates):
    for value in candidates:
        if value in values:
            break
    else:
        value = None
    return value


def get_data_var(data, variable):
    variable = find(data.data_vars, COMMON_VARIABLES[variable])
    if variable is None:
        raise ValueError(f"No data variable {variable} found")
    return variable


def force_minus_180_to_180(x):
    return (x + 180) % 360 - 180


def roll_from_0_360_to_minus_180_180(x):
    return np.argwhere(x[0] >= 180)[0][0]


def roll_from_minus_180_180_to_0_360(x):
    return np.argwhere(x[0] >= 0)[0][0]


def force_0_to_360(x):
    return x % 360


def extract_xy(
    field, x=None, y=None, src_crs=None, crs=None, bounds=None, data_vars=None
):
    points = field.to_points(flatten=False)
    values = field.to_numpy(flatten=False)

    if (
        bounds is not None
        and crs.__class__.__name__ not in domains.projections.FIXED_CRSS
    ):
        crs_bounds = domains.bounds.from_bbox(bounds, src_crs, crs)

        roll_by = None
        if src_crs.__class__.__name__ in domains.bounds.CYCLIC_SYSTEMS:
            if crs_bounds[0] < 0 and crs_bounds[1] > 0:
                if crs_bounds[0] < points["x"].min():
                    roll_by = roll_from_0_360_to_minus_180_180(points["x"])
                    points["x"] = force_minus_180_to_180(points["x"])
                    for i in range(2):
                        crs_bounds[i] = force_minus_180_to_180(crs_bounds[i])
            elif crs_bounds[0] < 180 and crs_bounds[1] > 180:
                if crs_bounds[1] > points["x"].max():
                    roll_by = roll_from_minus_180_180_to_0_360(points["x"])
                    points["x"] = force_0_to_360(points["x"])
                    for i in range(2):
                        crs_bounds[i] = force_0_to_360(crs_bounds[i])
        if roll_by is not None:
            points["x"] = np.roll(points["x"], roll_by, axis=1)
            points["y"] = np.roll(points["y"], roll_by, axis=1)
            values = np.roll(values, roll_by, axis=1)

        bbox = np.where(
            (points["x"] >= crs_bounds[0])
            & (points["x"] <= crs_bounds[1])
            & (points["y"] >= crs_bounds[2])
            & (points["y"] <= crs_bounds[3]),
            True,
            False,
        )

        kernel = np.ones((3, 3), dtype="uint8")
        bbox = cv2.dilate(bbox.astype("uint8"), kernel).astype(bool)

        shape = bbox[np.ix_(np.any(bbox, axis=1), np.any(bbox, axis=0))].shape

        points["x"] = points["x"][bbox].reshape(shape)
        points["y"] = points["y"][bbox].reshape(shape)
        values = values[bbox].reshape(shape)

    # try:
    #     for i, item in enumerate(values):
    #         values[i], points["x"] = add_cyclic_point(item, coord=points["x"])
    # except:  # noqa: E722
    #     # TODO: Decide what to do with un-cyclifiable data
    #     pass

    return values, points["x"], points["y"]


def _extract_axis(data, **kwargs):
    axis_name, axis_value = list(kwargs.items())[0]

    if axis_value is None:
        values = data.axis(axis_name).values
    else:
        values = data.to_xarray()[axis_value].values

    return values


def extract(data_vars=None):
    def wrapper(method):
        def sanitised_method(self, fields, *args, x=None, y=None, **kwargs):

            if not isinstance(fields, emohawk.sources.file.File):
                fields = [fields]

            # TODO: iteration over fields
            for field in fields[:1]:

                try:
                    proj_string = field.to_proj()[0]
                except AttributeError as err:
                    warnings.warn(f"{err}; assuming {schema.reference_crs} CRS")
                    crs = domains.crs.parse(schema.reference_crs)
                else:
                    if proj_string is None:
                        crs = domains.crs.parse(schema.reference_crs)
                    else:
                        crs = domains.projections.proj_to_ccrs(proj_string)

                if self._domain is None:
                    self._domain = domains.Domain(crs=crs)

                values, x, y = extract_xy(
                    field, x, y, crs, self.crs, self.bounds, data_vars=data_vars
                )

                kwargs["transform"] = kwargs.get("transform", crs)

                return method(self, values, *args, x=x, y=y, **kwargs)

        return sanitised_method

    return wrapper



# Copyright 2023, European Centre for Medium Range Weather Forecasts.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


def alias(**aliases):
    def decorator(function):
        def wrapper(*args, **kwargs):
            kwargs = {aliases.get(key, key): value for key, value in kwargs.items()}
            return function(*args, **kwargs)

        return wrapper

    return decorator



# Copyright 2023, European Centre for Medium Range Weather Forecasts.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from earthkit.maps import utils


def test_recursive_dict_update():
    dict_1 = {"nested": {"a": 1, "b": 2}, "c": 3, "d": 4}
    dict_2 = {"nested": {"a": 10, "c": 3}, "d": 5, "e": 6}

    expected = {"nested": {"a": 10, "b": 2, "c": 3}, "c": 3, "d": 5, "e": 6}

    assert utils.recursive_dict_update(dict_1, dict_2) == expected



# Copyright 2023, European Centre for Medium Range Weather Forecasts.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import cartopy.crs as ccrs
import cartopy.feature as cfeature
import matplotlib.pyplot as plt
from matplotlib.gridspec import GridSpec, SubplotSpec

from earthkit.maps import domains, inputs
from earthkit.maps.schema import schema

from . import _fonts, layers, markers, styles, titles  # noqa: F401


def await_crs(method):
    def wrapper(self, *args, **kwargs):
        if self._domain is None:
            self._queue.append((method, args, kwargs))
        else:
            return method(self, *args, **kwargs)

    return wrapper


class Chart:
    """
    A geospatial chart containing a single axis.

    Parameters
    ----------
    domain : str or list (optional)
        Must be either:
            - A string naming the domain to use for the chart (e.g. a country,
              continent or state/province).
            - A list containing the bounding box of the domain, in the order
              `[min_x, max_x, min_y, max_y]`. The bounding box is assumed to be
              defined in terms of latitude and longitude values, unless the
              `domain_crs` argument is also passed, in which case the bounding
              box must be defined using `domain_crs` coordinates.
        If no domain is passed, the chart will attempt to use the best domain
        to fit the first field of data plotted on the chart.
    crs : cartopy.crs.CRS (optional)
        The coordinate reference system to use for this chart. If `None`
        (default), will attempt to find the best CRS to suit the bounding box
        of the chart.
    domain_crs : cartopy.crs.CRS (optional)
        Use with `domain` to specify the coordinate system on which the domain
        bounding box is defined.
    figsize : list (optional)
        The width and height (in inches) of the chart. Note that the width of
        the chart may be adjusted slightly if the aspect ratio of your chart
        does not match the figsize.
    """

    def __init__(self, domain=None, crs=None, figsize=(10, 10)):
        self._domain = None
        if isinstance(domain, str):
            self._domain = domains.Domain.from_string(domain, crs)
        elif domain is not None or crs is not None:
            self._domain = domains.Domain(domain, crs)
        self._boundless = False

        self._queue = []

        self._fig = None
        self._figsize = figsize

        self._ax = None

        self._cbar_axes = None
        self._cbar_locations = None

        self._layers = []
        self._gridlines = None
        self._suptitle = None

    @staticmethod
    def close_current_fig():
        plt.clf()
        plt.close()

    @property
    def fig(self):
        if self._fig is None:
            self.close_current_fig()
            self._fig = plt.figure(figsize=self._figsize)
        return self._fig

    @property
    def ax(self):
        if self._ax is None:
            self._ax = self.fig.add_subplot(1, 1, 1, projection=self.crs)
            if self.bounds is not None:
                self._ax.set_extent(self.bounds, crs=self.crs)
        return self._ax

    @property
    def size(self):
        bbox = self.ax.get_position()
        return bbox.width, bbox.height

    def is_portrait(self):
        value = self.size[0] < self.size[1]
        return value

    @property
    def bounds(self):
        return self._domain.bounds

    @property
    def latlon_bounds(self):
        return self._domain.latlon_bounds

    @property
    def crs(self):
        return self._domain.crs

    def marker(self, *args, marker="o", transform=ccrs.Geodetic(), **kwargs):
        kwargs["transform"] = transform
        if isinstance(marker, markers.CustomMarker):
            plot = marker.plot(self, *args, **kwargs)
        else:
            plot = self.ax.plot(*args, marker=marker, **kwargs)
        return plot

    @styles.dynamic(normalize=True)
    @layers.append
    @inputs.extract()
    def mesh(self, data, *args, x=None, y=None, **kwargs):
        kwargs.pop("levels", None)
        return self.ax.pcolormesh(x, y, data, *args, **kwargs)

    @styles.dynamic(normalize=True)
    @layers.append
    @inputs.extract()
    def shaded_contour(
        self, data, *args, x=None, y=None, transform_first=True, **kwargs
    ):
        if self.crs.__class__.__name__ in domains.projections.FIXED_CRSS:
            transform_first = False
        return self.ax.contourf(
            x, y, data, *args, transform_first=transform_first, **kwargs
        )

    @layers.append
    @inputs.extract()
    @schema.contour.apply()
    def contour(
        self,
        data,
        *args,
        x=None,
        y=None,
        labels=False,
        label_fontsize=7,
        label_colors=None,
        label_frequency=1,
        label_background=None,
        transform_first=True,
        **kwargs,
    ):
        if self.crs.__class__.__name__ in domains.projections.FIXED_CRSS:
            transform_first = False
        if "cmap" in kwargs and "colors" in kwargs:
            kwargs.pop("colors")
        contours = self.ax.contour(
            x, y, data, *args, transform_first=transform_first, **kwargs
        )
        if labels:
            clabels = self.ax.clabel(
                contours,
                contours.levels[0::label_frequency],
                inline=True,
                fontsize=label_fontsize,
                colors=label_colors,
                inline_spacing=2,
            )
            if label_background is not None:
                for label in clabels:
                    label.set_backgroundcolor(label_background)
        return contours

    @schema.apply("font")
    @schema.legend.apply()
    def legend(self, layer=None, *args, **kwargs):
        if layer is None:
            return [self.legend(layer, *args, **kwargs) for layer in self._layers]
        elif isinstance(layer, int):
            layer = self._layers[layer]
        if layer.legend is not False:
            return layer.legend(*args, chart=self, **kwargs)

    def _resize_colorbars(self):
        p = self.ax.get_position()

        offset = {"right": 0, "left": 0, "bottom": 0, "top": 0}
        if self._gridlines is not None:
            pad = {"right": 0.045, "left": 0.045, "bottom": 0.02, "top": 0.02}
            for side in offset:
                if getattr(self._gridlines, f"{side}_labels"):
                    offset[side] += pad[side]

        for layer in self._layers:
            if layer._legend_location is None:
                continue
            position = {
                "right": [
                    p.x0 + p.width + 0.01 + offset["right"],
                    p.y0,
                    0.03,
                    p.height,
                ],
                "left": [p.x0 - 0.04 - offset["left"], p.y0, 0.03, p.height],
                "bottom": [p.x0, p.y0 - 0.04 - offset["bottom"], p.width, 0.03],
                "top": [p.x0, p.y0 + p.height + 0.04 + offset["top"], p.width, 0.03],
            }[layer._legend_location]
            offset[layer._legend_location] += 0.1
            layer._legend_ax.set_position(position)

        if self._suptitle:
            self._position_title()

    def _position_title(self):
        if any(layer._legend_location == "top" for layer in self._layers):
            y = (
                max(layer._legend_ax.get_position().ymax for layer in self._layers)
                + 0.06
            )
        else:
            y = self.ax.get_position().ymax + 0.025
            if self._gridlines is not None:
                if self._gridlines.top_labels:
                    y += 0.025
        x = self._suptitle.get_position()[0]
        self._suptitle.set_position((x, y))

    @await_crs
    @schema.coastlines.apply()
    def coastlines(self, **kwargs):
        """Add coastal outlines from the Natural Earth “coastline” shapefile collection."""
        return self.ax.coastlines(**kwargs)

    @await_crs
    @schema.borders.apply()
    def borders(self, **kwargs):
        """Add political boundaries from the Natural Earth administrative shapefile collection."""
        return self.ax.add_feature(cfeature.BORDERS, **kwargs)

    @await_crs
    @schema.roads.apply()
    def roads(self, resolution="50m", **kwargs):
        """Add roads from the Natural Earth cultural shapefile collection."""
        feature = cfeature.NaturalEarthFeature("cultural", "roads", "10m")

        return self.ax.add_feature(feature, **kwargs)

    @await_crs
    @schema.gridlines.apply()
    def gridlines(self, **kwargs):
        """Add latitude and longitude gridlines."""
        self._gridlines = self.ax.gridlines(**kwargs)

        # If the gridlines have labels, we need to update the colorbars
        if kwargs.get("draw_labels"):
            self._resize_colorbars()

        return self._gridlines

    @await_crs
    def land(self, **kwargs):
        """Add land polygons from the Natural Earth "land" shapefile collection."""
        return self.ax.add_feature(cfeature.LAND, **kwargs)

    @schema.apply("font")
    @schema.title.apply()
    def title(self, *args, **kwargs):
        if args:
            label = args[0]
            args = args[1:]
        else:
            label = kwargs.pop("labels", None)
        label = titles.format_string(self, label)

        plt.sca(self.ax)
        self._suptitle = plt.suptitle(label, *args, **kwargs)
        self._position_title()
        return self._suptitle

    def _release_queue(self):
        while len(self._queue):
            method, args, kwargs = self._queue.pop(0)
            method(self, *args, **kwargs)

    def show(self, *args, **kwargs):
        """Display the chart."""
        self._release_queue()
        plt.show(*args, **kwargs)

    def save(self, *args, **kwargs):
        """Save the chart."""
        self._release_queue()
        plt.savefig(*args, **kwargs)


class Tile(Chart):
    def __init__(self, *args, figsize=(512, 512), **kwargs):
        px = 1 / plt.rcParams["figure.dpi"]  # pixel in inches
        figsize = tuple([dim * px for dim in figsize])
        super().__init__(*args, figsize=figsize, **kwargs)

    @property
    def ax(self):
        if self._ax is None:
            subplotspec = SubplotSpec(
                GridSpec(1, 1, self.fig, left=0, right=1, bottom=0, top=1), 0
            )
            self._ax = self.fig.add_subplot(
                subplotspec, frameon=False, projection=self.crs
            )
            if self.bounds is not None:
                self._ax.set_extent(self.bounds, crs=self.crs)
            self._ax.get_xaxis().set_visible(False)
            self._ax.get_yaxis().set_visible(False)
        return self._ax

    def save(self, *args, **kwargs):
        self._release_queue()
        plt.savefig(*args, **kwargs)



# Copyright 2023, European Centre for Medium Range Weather Forecasts.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import cartopy.crs as ccrs
import pytest

from earthkit.maps import domains


def test_crs_from_string():
    crs = ccrs.PlateCarree()

    assert domains.crs_from_string("PlateCarree") == crs

    with pytest.raises(ValueError) as e_info:  # noqa: F841
        domains.crs_from_string("foobarbaz")


def test_crs_from_dict():
    crs = ccrs.PlateCarree(central_longitude=50)

    assert (
        domains.crs_from_dict({"name": "PlateCarree", "central_longitude": 50}) == crs
    )


def test_parse_crs():
    crs = ccrs.PlateCarree()
    assert domains.parse_crs("PlateCarree") == crs

    crs = ccrs.PlateCarree(central_longitude=50)
    assert domains.parse_crs({"name": "PlateCarree", "central_longitude": 50}) == crs

    crs = domains.DEFAULT_CRS
    assert domains.parse_crs(None) == crs



# Copyright 2023, European Centre for Medium Range Weather Forecasts.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import cartopy.crs as ccrs
import pytest

from earthkit.maps.domains import auto


def test_preset():
    expected_bounds = [2502500, 7497500, 752500, 5497500]
    expected_crs = ccrs.LambertAzimuthalEqualArea(
        central_latitude=52,
        central_longitude=10,
        false_easting=4321000,
        false_northing=3210000,
    )

    bounds, crs = auto.get_bounds_and_crs("europe")

    assert bounds == expected_bounds
    assert crs == expected_crs


def test_country():
    expected_bounds = [-626161, 690812, -625183, 649633]
    expected_crs_params = {
        "ellps": "WGS84",
        "proj": "aea",
        "lon_0": 2.5,
        "lat_0": 46.3,
        "x_0": 0.0,
        "y_0": 0.0,
        "lat_1": 42.9,
        "lat_2": 49.6,
    }

    bounds, crs = auto.get_bounds_and_crs("france")

    assert bounds == pytest.approx(expected_bounds, 0.5)

    crs_params = crs.proj4_params
    for key in expected_crs_params:
        if isinstance(crs_params[key], (int, float)):
            assert crs_params[key] == pytest.approx(expected_crs_params[key], 1)
        else:
            assert crs_params[key] == expected_crs_params[key]


def test_alias():
    expected_bounds = [-363796, 373425, -541558, 545791]
    expected_crs_params = {
        "ellps": "WGS84",
        "proj": "aea",
        "lon_0": -2.9,
        "lat_0": 54.3,
        "x_0": 0.0,
        "y_0": 0.0,
        "lat_1": 51.3,
        "lat_2": 57.2,
    }

    bounds, crs = auto.get_bounds_and_crs("UK")

    assert bounds == pytest.approx(expected_bounds, 0.5)

    crs_params = crs.proj4_params
    for key in expected_crs_params:
        if isinstance(crs_params[key], (int, float)):
            assert crs_params[key] == pytest.approx(expected_crs_params[key], 1)
        else:
            assert crs_params[key] == expected_crs_params[key]


def test_get_optimal_crs():
    bounds = [-25, 40, 32, 72]
    expected_crs_params = {
        "ellps": "WGS84",
        "proj": "aea",
        "lon_0": 7.5,
        "lat_0": 52.0,
        "x_0": 0.0,
        "y_0": 0.0,
        "lat_1": 38.4,
        "lat_2": 65.6,
    }

    crs = auto.get_optimal_crs(bounds)

    assert crs.proj4_params == expected_crs_params


def test_get_crs_extents():
    bounds = [-25, 40, 32, 72]
    crs = ccrs.AlbersEqualArea(
        central_latitude=52, central_longitude=7.5, standard_parallels=(38.4, 65.6)
    )

    expected_bounds = [-3046054.6, 3046054.6, -2239135.9, 2497907.6]

    result = auto.get_crs_extents(bounds, crs)

    assert result == pytest.approx(expected_bounds, 0.5)


def test_Extents_is_global():
    # Test whole globe
    domain = [-180, 180, -90, 90]
    extents = auto.Extents(domain)
    assert extents.is_global()

    # Test > 60% coverage but < 100%
    domain = [-150, 150, -65, 65]
    extents = auto.Extents(domain)
    assert extents.is_global()

    # Test < 60% coverage
    domain = [-140, 140, -65, 65]
    extents = auto.Extents(domain)
    assert not extents.is_global()

    # Test large threshold
    domain = [-150, 150, -65, 65]
    extents = auto.Extents(domain, large_threshold=0.7)
    assert not extents.is_global()


def test_Extents_is_equatorial():
    # Test whole globe
    domain = [-180, 180, -90, 90]
    extents = auto.Extents(domain)
    assert extents.is_equatorial()

    # Test region centred on equator
    domain = [-10, 40, -30, 30]
    extents = auto.Extents(domain)
    assert extents.is_equatorial()

    # Test region close to equator
    domain = [-30, 20, -15, 45]
    extents = auto.Extents(domain)
    assert extents.is_equatorial()

    # Test region on boundary
    domain = [-30, 20, 20, 30]
    extents = auto.Extents(domain)
    assert not extents.is_equatorial()

    # Test non-equatorial region
    domain = [-30, 20, 60, 70]
    extents = auto.Extents(domain)
    assert not extents.is_equatorial()


def test_Extents_is_polar():
    # Test whole globe
    domain = [-180, 180, -90, 90]
    extents = auto.Extents(domain)
    assert not extents.is_polar()

    # Test region centred on north pole
    domain = [-180, 180, 70, 90]
    extents = auto.Extents(domain)
    assert extents.is_polar()

    # Test region close to north pole
    domain = [-180, 180, 75, 85]
    extents = auto.Extents(domain)
    assert extents.is_polar()

    # Test region centred on south pole
    domain = [-180, 180, -90, -70]
    extents = auto.Extents(domain)
    assert extents.is_polar()

    # Test region on boundary
    domain = [-180, 180, 60, 90]
    extents = auto.Extents(domain)
    assert not extents.is_polar()

    # Test non-polar region
    domain = [-10, 40, 30, 70]
    extents = auto.Extents(domain)
    assert not extents.is_polar()


def test_Extents_is_portrait_landscape_square():
    # Test portrait
    domain = [-10, 20, 30, 90]
    extents = auto.Extents(domain)
    assert extents.is_portrait()
    assert not extents.is_landscape()
    assert not extents.is_square()

    # Test just above portrait boundary
    domain = [-10, 20, 30, 69]
    extents = auto.Extents(domain)
    assert extents.is_portrait()
    assert not extents.is_landscape()
    assert not extents.is_square()

    # Test on portrait boundary
    domain = [-10, 20, 30, 66]
    extents = auto.Extents(domain)
    assert not extents.is_portrait()
    assert not extents.is_landscape()
    assert extents.is_square()

    # Test landscape
    domain = [-30, 20, 30, 60]
    extents = auto.Extents(domain)
    assert not extents.is_portrait()
    assert extents.is_landscape()
    assert not extents.is_square()

    # Test just above landscape boundary
    domain = [-19, 20, 30, 60]
    extents = auto.Extents(domain)
    assert not extents.is_portrait()
    assert extents.is_landscape()
    assert not extents.is_square()

    # Test on landscape boundary
    domain = [-16, 20, 30, 66]
    extents = auto.Extents(domain)
    assert not extents.is_portrait()
    assert not extents.is_landscape()
    assert extents.is_square()

    # Test square
    domain = [-10, 20, 30, 60]
    extents = auto.Extents(domain)
    assert not extents.is_portrait()
    assert not extents.is_landscape()
    assert extents.is_square()


def test_Extents_is_large_small():
    # Test whole globe
    domain = [-180, 180, -90, 90]
    extents = auto.Extents(domain)
    assert not extents.is_large()
    assert not extents.is_small()

    # Test > 60% coverage but < 100%
    domain = [-150, 150, -65, 65]
    extents = auto.Extents(domain)
    assert not extents.is_large()
    assert not extents.is_small()

    # Test < 60% coverage
    domain = [-140, 140, -65, 65]
    extents = auto.Extents(domain)
    assert extents.is_large()
    assert not extents.is_small()

    # Test > 20% coverage byt < 60%
    domain = [-100, 100, -40, 30]
    extents = auto.Extents(domain)
    assert extents.is_large()
    assert not extents.is_small()

    # Test 20% coverage
    domain = [-80, 80, -40, 40]
    extents = auto.Extents(domain)
    assert not extents.is_large()
    assert extents.is_small()



# Copyright 2023, European Centre for Medium Range Weather Forecasts.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import warnings

import cartopy.crs as ccrs
import pyproj

FIXED_CRSS = ["NorthPolarStereo", "SouthPolarStereo"]


PROJ4_CRS = {
    "laea": {
        "crs": ccrs.LambertAzimuthalEqualArea,
        "kwargs": {
            "central_longitude": "lon_0",
            "central_latitude": "lat_0",
            "false_easting": "x_0",
            "false_northing": "y_0",
        },
    }
}


def proj_to_ccrs(proj4_string):

    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        params = pyproj.CRS.from_proj4(proj4_string).to_dict()

    ccrs_setup = PROJ4_CRS[params["proj"]]
    crs = ccrs_setup["crs"]
    kwargs = ccrs_setup["kwargs"]

    return crs(**{kwarg: params[kwargs[kwarg]] for kwarg in kwargs})



# Copyright 2023, European Centre for Medium Range Weather Forecasts.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from earthkit.maps.charts import Chart


class Subplots(Chart):
    def __init__(self, *args, rows=None, cols=None, **kwargs):
        super().__init__(*args, **kwargs)

        self._rows = rows
        self._cols = cols

    @property
    def ax(self):
        if self._ax is None:
            self._ax = self.fig.add_subplot(
                self._rows, self._cols, 1, projection=self.crs
            )
            if self.bounds is not None:
                self._ax.set_extent(self.bounds, crs=self.crs)
        return self._ax



# Copyright 2023, European Centre for Medium Range Weather Forecasts.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from string import Formatter

import emohawk
from pandas import Timestamp

NO_CFUNITS = False
try:
    from cf_units.tex import tex
except ImportError:
    NO_CFUNITS = True

METADATA = {
    "variable_name": {
        "preference": ["long_name", "standard_name", "name", "short_name"],
    }
}


UNCAPITALIZED = ["a", "is", "of", "the"]


class _DummyChart:
    def __init__(self, layer):
        self._layers = [layer]


class TitleFormatter(Formatter):

    TOP_LEVEL_KEYS = {
        "domain": "_domain",
    }

    def __init__(self, chart):
        super().__init__()
        self._chart = chart
        self._current_layer = None

    def format(self, format_string, /, *args, **kwargs):
        format_keys = [i[1] for i in self.parse(format_string) if i[1] is not None]
        for key in format_keys:
            if key not in kwargs:
                kwargs[key] = key
        return super().format(format_string, *args, **kwargs)

    def convert_field(self, value, conversion):
        if conversion is not None and conversion.isnumeric():
            self._current_layer = int(conversion)
            conversion = None
        return super().convert_field(value, conversion)

    def format_field(self, value, format_spec):
        key = value
        if value in self.TOP_LEVEL_KEYS:
            value = getattr(self._chart, self.TOP_LEVEL_KEYS[value])
        else:
            value = get_metadata(self._chart._layers, value, self._current_layer)
        if value == key:
            value = f"no_{value}".upper()
        f = super().format_field
        if isinstance(value, list):
            result = list_to_human(f(v, format_spec) for v in value)
        else:
            result = f(value, format_spec)
        self._current_layer = None
        return result


def get_metadata(layers, attr, layer=None):
    if layer is not None:
        layers = [layers[layer]]

    labels = []
    for layer in layers:

        if attr == "time":
            label = layer.data.to_datetime()
        else:
            if not type(layer.data[0]).__name__ == "GribField":
                label = xarray_metadata(layer.data, attr)
            else:
                candidates = [attr]
                if attr in METADATA:
                    candidates = METADATA[attr]["preference"] + candidates

                for item in candidates:
                    label = layer.data.metadata(item)[0]
                    if label is not None:
                        break
                else:
                    label = f"<NO {attr.upper()}>"

        if isinstance(label, str) and attr == "units" and not NO_CFUNITS:
            label = f"${tex(label)}$"

        labels.append(label)

    return labels


def xarray_metadata(data, attr):
    data = data[0].to_xarray()

    preference = [attr]
    if attr in METADATA:
        preference = METADATA[attr]["preference"] + preference

    for item in preference:
        try:
            label = getattr(data, item)
            break
        except AttributeError:
            continue
    else:
        label = attr

    try:
        label = label.values
    except AttributeError:
        pass
    else:
        try:
            if len(label) == 1:
                label = label.item()
        except TypeError:
            pass
        label = Timestamp(label).to_pydatetime()

    return label


def format_string(self, label):
    if label is None:
        return _default_title(self)
    formatter = TitleFormatter(self)
    return formatter.format(label)


def title_string(string):
    working_string = string.replace("_", " ").title()
    chunks = working_string.split()
    for i in range(len(chunks)):
        if chunks[i].lower() in UNCAPITALIZED:
            chunks[i] = chunks[i].lower()
    chunks[0] = chunks[0].capitalize()
    return " ".join(chunks)


def _default_title(self):
    times = [
        emohawk.from_source("file", layer.data).to_datetime() for layer in self._layers
    ]
    if len(set(times)) == 1:
        label = "{variable_name} at {time!0:%H:%M} on {time!0:%Y-%m-%d}"
    else:
        label = list_to_human(
            f"{{variable_name!{i}}} at {{time!{i}:%H:%M}} on {{time!{i}:%Y-%m-%d}}"
            for i in range(len(self._layers))
        )
    if self.bounds is not None:
        label += f" over {self._domain.title}"
    return format_string(self, label)


def colorbar_title(label, layer):
    chart = _DummyChart(layer)
    formatter = TitleFormatter(chart)

    if label is None:
        label = "{variable_name} ({units})"

    return formatter.format(label)


def list_to_human(iterable, conjunction="and", oxford_comma=False):
    """Convert a list to a human-readable string."""
    list_of_strs = [str(item) for item in iterable]

    if len(list_of_strs) > 2:
        list_of_strs = [", ".join(list_of_strs[:-1]), list_of_strs[-1]]
        if oxford_comma:
            list_of_strs[0] += ","

    return f" {conjunction} ".join(list_of_strs)



# Copyright 2023, European Centre for Medium Range Weather Forecasts.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import cartopy.crs as ccrs

from earthkit.maps import _data, domains

DOMAIN_LOOKUP = _data.load("domains")


class Domain:
    @classmethod
    def from_string(cls, string, crs=None):
        from . import natural_earth

        domain_name = lookup_name(string)

        if domain_name is not None and domain_name in DOMAIN_LOOKUP["domains"]:
            domain_config = DOMAIN_LOOKUP["domains"][domain_name]
            if isinstance(domain_config, list):
                bounds = domain_config
                domain_crs = None
            else:
                bounds = domain_config.get("bounds")
                domain_crs = domain_config.get("crs")
            if crs is not None:
                bounds = domains.bounds.from_bbox(
                    bounds, crs, domains.crs.parse(domain_crs)
                )
            else:
                crs = domains.crs.parse(
                    domain_crs or domains.optimal.crs_from_bounds(bounds)
                )
                if domain_crs is None:
                    crs = domains.optimal.crs_from_bounds(bounds)
                    bounds = domains.bounds.from_bbox(bounds, crs)

        else:
            domain_name = domain_name or string
            source = natural_earth.NaturalEarthDomain(domain_name, crs)
            domain_name = source.domain_name
            bounds = source.bounds
            crs = source.crs

        return cls(bounds, crs, domain_name)

    def __init__(self, bounds=None, crs=None, domain_name=None):
        if crs is None:
            self.crs = domains.optimal.crs_from_bounds(bounds)
            self.bounds = domains.bounds.from_bbox(bounds, self.crs)
        else:
            self.crs = domains.crs.parse(crs)
            self.bounds = bounds

        self.domain_name = domain_name

    @property
    def title(self):
        if self.domain_name in DOMAIN_LOOKUP["the_countries"]:
            return f"the {self.domain_name}"
        elif self.domain_name is None:
            string = "a custom domain"
            if self.bounds is not None:
                string += f" ({[round(i, 2) for i in self.latlon_bounds]})"
            return string
        return self.domain_name

    @property
    def latlon_bounds(self):
        return domains.bounds.from_bbox(self.bounds, ccrs.PlateCarree(), self.crs)


def lookup_name(domain_name):
    """Format a domain name."""
    # normalise the input string and the lookup key
    domain_name = domain_name.lower().replace("_", " ")
    name_mapping = {k.lower(): k for k in DOMAIN_LOOKUP["domains"]}

    if domain_name not in name_mapping:
        for name, alt_names in DOMAIN_LOOKUP["alternate_names"].items():
            if domain_name in [alt_name.lower() for alt_name in alt_names]:
                domain_name = name
                break
        else:
            domain_name = None
    domain_name = name_mapping.get(domain_name, domain_name)

    return domain_name



# Copyright 2023, European Centre for Medium Range Weather Forecasts.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from functools import partial
from pathlib import Path

import cartopy
import matplotlib
import pytest

from earthkit.maps import charts, schema


# TODO: There is a bug in matplotlib causing an incorrect DeprecationWarning
# to be raised; ignore for now, but update when matplotlib has fixed this bug
@pytest.mark.filterwarnings("ignore::DeprecationWarning")
def test_Chart_queue():
    chart = charts.Chart()

    assert len(chart._queue) == 0

    chart.coastlines()
    assert len(chart._queue) == 1
    assert chart._queue[0][1] == tuple()
    assert chart._queue[0][2] == dict()

    chart._release_queue()
    assert len(chart._queue) == 0


@pytest.mark.filterwarnings("ignore::DeprecationWarning")
def test_Chart_bounds_and_crs():
    chart = charts.Chart()
    assert chart.bounds is None
    assert chart.crs == schema.reference_crs

    expected_bounds = [-363796, 373425, -541558, 545791]
    expected_crs_params = {
        "ellps": "WGS84",
        "proj": "aea",
        "lon_0": -2.9,
        "lat_0": 54.3,
        "x_0": 0.0,
        "y_0": 0.0,
        "lat_1": 51.3,
        "lat_2": 57.2,
    }
    chart = charts.Chart(domain="United Kingdom")
    assert chart.bounds == pytest.approx(expected_bounds, 0.5)
    crs_params = chart.crs.proj4_params
    for key in expected_crs_params:
        if isinstance(crs_params[key], (int, float)):
            assert crs_params[key] == pytest.approx(expected_crs_params[key], 1)
        else:
            assert crs_params[key] == expected_crs_params[key]


@pytest.fixture
def dummy_chart():
    path = Path(__file__).parent / "temporary-chart.png"
    chart = charts.Chart()
    chart.temporary_save = partial(chart.save, path)
    yield chart
    os.remove(path)


@pytest.mark.filterwarnings("ignore::DeprecationWarning")
def test_Chart_coastlines(dummy_chart):
    basic_children = dummy_chart.axis.get_children()

    dummy_chart.coastlines()
    dummy_chart.temporary_save()
    new_children = dummy_chart.axis.get_children()
    assert len(new_children) == len(basic_children) + 1
    assert isinstance(new_children[0], cartopy.mpl.feature_artist.FeatureArtist)


@pytest.mark.filterwarnings("ignore::DeprecationWarning")
def test_Chart_gridlines(dummy_chart):
    basic_children = dummy_chart.axis.get_children()

    dummy_chart.gridlines()
    dummy_chart.temporary_save()
    new_children = dummy_chart.axis.get_children()
    assert len(new_children) > len(basic_children)
    assert isinstance(new_children[0], matplotlib.collections.LineCollection)


@pytest.mark.filterwarnings("ignore::DeprecationWarning")
def test_Chart_borders(dummy_chart):
    basic_children = dummy_chart.axis.get_children()

    dummy_chart.borders()
    dummy_chart.temporary_save()
    new_children = dummy_chart.axis.get_children()
    assert len(new_children) == len(basic_children) + 1
    assert isinstance(new_children[0], cartopy.mpl.feature_artist.FeatureArtist)


@pytest.mark.filterwarnings("ignore::DeprecationWarning")
def test_Chart_land(dummy_chart):
    basic_children = dummy_chart.axis.get_children()

    dummy_chart.land()
    dummy_chart.temporary_save()
    new_children = dummy_chart.axis.get_children()
    assert len(new_children) == len(basic_children) + 1
    assert isinstance(new_children[0], cartopy.mpl.feature_artist.FeatureArtist)



# Copyright 2023, European Centre for Medium Range Weather Forecasts.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import emohawk


class DataLayer:

    from .legends import legend

    def __init__(self, data, layer, legend=None):
        self.data = data
        self.layer = layer
        self._legend = legend
        self._legend_location = None
        self._legend_ax = None


def append(method):
    def wrapper(self, data, *args, legend=True, **kwargs):
        if isinstance(data, str):
            data = emohawk.from_source("file", data)
        layer = method(self, data, *args, **kwargs)
        self._layers.append(DataLayer(data, layer, legend))
        self._release_queue()
        return layer

    return wrapper



# Copyright 2023, European Centre for Medium Range Weather Forecasts.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import matplotlib.pyplot as plt

from . import titles


def continuous_colorbar(self, args, axis, orientation, kwargs):
    colorbar = plt.colorbar(
        self.layer,
        *args,
        orientation=orientation,
        cax=axis,
        **kwargs,
    )
    colorbar.ax.minorticks_off()
    return colorbar


def discrete(axis):
    pass


def categorical(axis):
    pass


LEGEND_MODES = {
    "continuous": continuous_colorbar,
    "discrete": discrete,
    "categorical": categorical,
}


FALLBACK_LEGEND = continuous_colorbar


def _default_legend(layer_type):
    from cartopy.mpl import contour, geocollection

    return {
        geocollection.GeoQuadMesh: continuous_colorbar,
        contour.GeoContourSet: continuous_colorbar,
    }.get(layer_type, FALLBACK_LEGEND)


def legend(
    data_layer,
    *args,
    chart,
    mode=None,
    orientation=None,
    location=None,
    title=None,
    font=None,
    **kwargs,
):
    if data_layer._legend is False:
        return

    if orientation and location:
        raise ValueError(
            "'orientation' and 'location' are mutually exclusive arguments"
        )
    elif orientation is not None:
        location = "right" if orientation == "vertical" else "bottom"
    elif location is None:
        # Default to a colorbar on the right if the plot is portrait, or the
        # bottom if it's landscape
        location = "right" if chart.is_portrait() else "bottom"
    data_layer._legend_location = location
    orientation = "vertical" if location in ("left", "right") else "horizontal"

    # Create an arbitrary axis; this will be moved by mpl tight_layout
    data_layer._legend_ax = chart.fig.add_axes([0, 0, 0.1, 0.1])
    chart.fig.canvas.mpl_connect("resize_event", chart._resize_colorbars)

    if mode is None:
        draw_legend = _default_legend(type(data_layer.layer))
    else:
        if mode in LEGEND_MODES:
            draw_legend = LEGEND_MODES[mode]
        else:
            raise ValueError(
                f"invalid mode {mode}; must be one of "
                f"{','.join(list(LEGEND_MODES))}"
            )

    data_layer._legend = draw_legend(
        data_layer, args, data_layer._legend_ax, orientation, kwargs
    )
    if location == "left":
        data_layer._legend_ax.yaxis.set_ticks_position("left")

    rotation = {"right": 270, "left": 90, "top": 0, "bottom": 0}[location]
    labelpad = {"right": 20, "left": -60, "top": -60, "bottom": 0}[location]
    data_layer._legend.set_label(
        titles.colorbar_title(title, data_layer),
        rotation=rotation,
        labelpad=labelpad,
        font=font,
    )

    ticks = {
        "vertical": data_layer._legend.ax.get_yticklabels(),
        "horizontal": data_layer._legend.ax.get_xticklabels(),
    }[orientation]
    [tick.set_font(font) for tick in ticks]

    chart._resize_colorbars()



# Copyright 2023, European Centre for Medium Range Weather Forecasts.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from earthkit.maps import schema


def test_default_schema():
    schema.use("default")


def test_use():
    schema.use("light")
    assert schema.colormap == "viridis"


def test_set():
    font = "verdana"
    schema.set(font=font)
    assert schema.font == font


def test_set_nested():
    labels = "foobarbaz"
    schema.contour.set(labels=labels)
    assert schema.contour.labels == labels


def test_set_context_manager():
    font = "verdana"
    schema.set(font=font)

    tmp_font = "comic sans"
    with schema.set(font=tmp_font):
        assert schema.font == tmp_font

    assert schema.font == font



# Copyright 2023, European Centre for Medium Range Weather Forecasts.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class CustomMarker:
    pass


class oktas(CustomMarker):

    DEFAULT_KWARGS = {
        "color": "white",
        "markeredgecolor": "#333333",
        "markeredgewidth": 0.75,
        "markersize": 10,
    }

    def __init__(self, oktas, **kwargs):
        self.oktas = oktas
        super().__init__(**kwargs)

    def plot(self, chart, *args, **kwargs):
        kwargs = {**self.DEFAULT_KWARGS, **kwargs}
        return {
            0: _zero_oktas,
            1: _one_oktas,
            2: _two_oktas,
            3: _three_oktas,
            4: _four_oktas,
            5: _five_oktas,
            6: _six_oktas,
            7: _seven_oktas,
            8: _eight_oktas,
            9: _nine_oktas,
        }[self.oktas](chart, *args, **kwargs)


def _zero_oktas(chart, *args, **kwargs):
    return (chart.ax.plot(*args, marker="o", **kwargs),)


def _one_oktas(chart, *args, **kwargs):
    return (
        chart.ax.plot(
            *args,
            marker="o",
            markerfacecoloralt=kwargs["color"],
            fillstyle="left",
            **kwargs,
        ),
    )


def _two_oktas(chart, *args, **kwargs):
    layer_0 = chart.ax.plot(
        *args,
        marker="o",
        markerfacecoloralt=kwargs["markeredgecolor"],
        fillstyle="left",
        **{**kwargs, **{"markeredgecolor": "none"}},
    )
    layer_1 = chart.ax.plot(
        *args,
        marker="o",
        markerfacecoloralt="none",
        fillstyle="bottom",
        **{**kwargs, **{"markeredgecolor": "none"}},
    )
    layer_2 = chart.ax.plot(
        *args,
        marker="o",
        **{**kwargs, **{"color": "none"}},
    )
    return (
        layer_0,
        layer_1,
        layer_2,
    )


def _three_oktas(chart, *args, **kwargs):
    layer_0 = chart.ax.plot(
        *args,
        marker="o",
        markerfacecoloralt=kwargs["markeredgecolor"],
        fillstyle="left",
        **{**kwargs, **{"markeredgecolor": "none"}},
    )
    layer_1 = chart.ax.plot(
        *args,
        marker="o",
        markerfacecoloralt="none",
        fillstyle="bottom",
        **{**kwargs, **{"markeredgecolor": "none"}},
    )
    layer_2 = chart.ax.plot(
        *args,
        marker="o",
        markerfacecoloralt="none",
        fillstyle="left",
        **{**kwargs, **{"color": "none"}},
    )
    return (
        layer_0,
        layer_1,
        layer_2,
    )


def _four_oktas(chart, *args, **kwargs):
    return (
        chart.ax.plot(
            *args,
            marker="o",
            markerfacecoloralt=kwargs["markeredgecolor"],
            fillstyle="left",
            **kwargs,
        ),
    )


def _five_oktas(chart, *args, **kwargs):
    layer_0 = _four_oktas(chart, *args, **kwargs)[0]
    layer_1 = chart.ax.plot(
        *args,
        marker="o",
        markerfacecoloralt="none",
        fillstyle="top",
        **{**kwargs, **{"color": "none"}},
    )
    return (
        layer_0,
        layer_1,
    )


def _six_oktas(chart, *args, **kwargs):
    layer_0 = chart.ax.plot(
        *args,
        marker="o",
        markerfacecoloralt=kwargs["markeredgecolor"],
        fillstyle="left",
        **{**kwargs, **{"markeredgecolor": "none"}},
    )
    layer_1 = chart.ax.plot(
        *args,
        marker="o",
        markerfacecoloralt=kwargs["markeredgecolor"],
        fillstyle="top",
        **{**kwargs, **{"color": "none", "markeredgecolor": "none"}},
    )
    layer_2 = chart.ax.plot(*args, marker="o", **{**kwargs, **{"color": "none"}})
    return layer_0, layer_1, layer_2


def _seven_oktas(chart, *args, **kwargs):
    layer_0 = _eight_oktas(chart, *args, **kwargs)[0]
    layer_1 = chart.ax.plot(
        *args,
        marker="|",
        **{
            **kwargs,
            **{
                "markeredgecolor": kwargs["color"],
                "markeredgewidth": kwargs["markeredgewidth"] * 3,
            },
        },
    )
    layer_2 = chart.ax.plot(*args, marker="o", **{**kwargs, **{"color": "none"}})
    return layer_0, layer_1, layer_2


def _eight_oktas(chart, *args, **kwargs):
    return (
        chart.ax.plot(
            *args, marker="o", **{**kwargs, **{"color": kwargs["markeredgecolor"]}}
        ),
    )


def _nine_oktas(chart, *args, **kwargs):
    layer_0 = _zero_oktas(chart, *args, **kwargs)[0]
    layer_1 = chart.ax.plot(
        *args, marker="x", **{**kwargs, **{"markersize": kwargs["markersize"] * 0.71}}
    )
    return layer_0, layer_1
